from typing import List, Union

from mindsdb_sql.parser.dialects.mindsdb import CreateDatabase
from mindsdb_sql.parser.ast import DropDatabase, Identifier

from mindsdb_sdk.utils.objects_collection import CollectionBase

from .query import Query
from .tables import Tables
from .handlers import Handler


class Database:
    """
    Allows to work with database (datasource): to use tables and make raw queries

    To run native query
    At this moment query is just saved in Qeury object and not executed

    >>> query = database.query('select * from table1') # returns Query

    This command sends request to server to execute query and return dataframe

    >>> df = query.fetch()

    Has list of tables in .tables attribute.

    """

    def __init__(self, server, name, engine=None):
        self.server = server
        self.name = name
        self.engine = engine
        self.api = server.api

        self.tables = Tables(self, self.api)

        # old api
        self.get_table = self.tables.get
        self.list_tables = self.tables.list
        self.create_table = self.tables.create
        self.drop_table = self.tables.drop

    def __repr__(self):
        return f'{self.__class__.__name__}({self.name})'

    def query(self, sql: str) -> Query:
        """
        Make raw query to integration

        :param sql: sql of the query
        :return: Query object
        """
        return Query(self.api, sql, database=self.name)


class Databases(CollectionBase):
    """
    Databases
    ----------

    >>> databases.list()
    >>> db = databases[0] # Database type object

    # create

    >>> db = databases.create('example_db',
    ...                                 type='postgres',
    ...                                 connection_args={'host': ''})

    # drop database

    >>> databases.drop('example_db')

    # get existing

    >>> db = databases.get('example_db')

    """

    def __init__(self, api):
        self.api = api

    def _list_databases(self):
        data = self.api.sql_query(
            "select NAME, ENGINE from information_schema.databases where TYPE='data'"
        )
        return dict(zip(data.NAME, data.ENGINE))

    def list(self) -> List[Database]:
        """
        Show list of integrations (databases) on server

        :return: list of Database objects
        """
        databases = self._list_databases()
        return [Database(self, name, engine=engine) for name, engine in databases.items()]

    def create(self, name: str, engine: Union[str, Handler], connection_args: dict) -> Database:
        """
        Create new integration and return it

        :param name: Identifier for the integration to be created
        :param engine: Engine to be selected depending on the database connection.
        :param connection_args: {"key": "value"} object with the connection parameters specific for each engine
        :return: created Database object
        """
        if isinstance(engine, Handler):
            engine = engine.name

        ast_query = CreateDatabase(
            name=Identifier(name),
            engine=engine,
            parameters=connection_args,
        )
        self.api.sql_query(ast_query.to_string())
        return Database(self, name, engine=engine)

    def drop(self, name: str):
        """
        Delete integration

        :param name: name of integration
        """
        ast_query = DropDatabase(name=Identifier(name))
        self.api.sql_query(ast_query.to_string())

    def get(self, name: str) -> Database:
        """
        Get integration by name

        :param name: name of integration
        :return: Database object
        """
        databases = self._list_databases()
        if name not in databases:
            raise AttributeError("Database doesn't exist")
        return Database(self, name, engine=databases[name])



from typing import  List

from mindsdb_sql.parser.dialects.mindsdb import CreateDatabase, DropPredictor
from mindsdb_sql.parser.ast import DropDatabase
from mindsdb_sql.parser.ast import Identifier

from mindsdb_sdk.agents import Agents
from mindsdb_sdk.databases import Databases
from mindsdb_sdk.ml_engines import MLEngines
from mindsdb_sdk.skills import Skills
from mindsdb_sdk.utils.objects_collection import CollectionBase

from .models import Models
from .query import Query
from .views import Views
from .jobs import Jobs
from .knowledge_bases import KnowledgeBases


class Project:
    """
    Allows to work with project: to manage models and views inside of it or call raw queries inside of project

    Server instance allows to manipulate project and databases (integration) on mindsdb server

    Attributes for accessing to different objects:
        - models, see :func:`~mindsdb_sdk.models.Models`
        - views, see :func:`~mindsdb_sdk.views.Views`
        - jobs, see :func:`~mindsdb_sdk.jobs.Jobs`

    It is possible to cal queries from project context:

    Making prediciton using sql:

    >>> query = project.query('select * from database.table join model1')
    >>> df = query.fetch()

    Making time series prediction:

    >>> df = project.query('''
    ...      SELECT m.saledate as date, m.ma as forecast
    ...     FROM mindsdb.house_sales_model as m
    ...     JOIN example_db.demo_data.house_sales as t
    ...     WHERE t.saledate > LATEST AND t.type = 'house'
    ...     AND t.bedrooms=2
    ...     LIMIT 4;
    ...    ''').fetch()

    """

    def __init__(self, api, name, agents: Agents = None, skills: Skills = None, knowledge_bases: KnowledgeBases = None, databases: Databases = None, ml_engines: MLEngines = None):
        self.name = name
        self.api = api

        self.models = Models(self, api)

        # old api
        self.get_model = self.models.get
        self.list_models = self.models.list
        self.create_model = self.models.create
        self.drop_model = self.models.drop

        self.views = Views(self, api)

        # old api
        self.get_view = self.views.get
        self.list_views = self.views.list
        self.create_view = self.views.create
        self.drop_view = self.views.drop

        self.jobs = Jobs(self, api)

        # old api
        self.get_job = self.jobs.get
        self.list_jobs = self.jobs.list
        self.create_job = self.jobs.create
        self.drop_job = self.jobs.drop

        self.databases = databases or Databases(api)
        self.knowledge_bases = knowledge_bases or KnowledgeBases(self, api)

        self.skills = skills or Skills(api, name)
        self.agents = agents or Agents(api, name, self.knowledge_bases, self.databases, self.models, ml_engines, self.skills)

    def __repr__(self):
        return f'{self.__class__.__name__}({self.name})'

    def query(self, sql: str) -> Query:
        """
        Execute raw query inside of project

        :param sql: sql query
        :return: Query object
        """
        return Query(self.api, sql, database=self.name)

    def drop_model_version(self, name: str, version: int):
        """
        Drop version of the model

        :param name: name of the model
        :param version: version to drop
        """
        ast_query = DropPredictor(Identifier(parts=[name, str(version)]))

        self.query(ast_query.to_string()).fetch()



class Projects(CollectionBase):
    """
    Projects
    ----------

    list of projects

    >>> projects.list()

    create

    >>> project = projects.create('proj')

    drop

    >>> projects.drop('proj')

    get existing

    >>> projects.get('proj')

    by attribute
    >>> projects.proj

    """

    def __init__(self, api):
        self.api = api

    def _list_projects(self):
        data = self.api.sql_query("select NAME from information_schema.databases where TYPE='project'")
        return list(data.NAME)

    def list(self) -> List[Project]:
        """
        Show list of project on server

        :return: list of Project objects
        """
        # select * from information_schema.databases where TYPE='project'
        return [Project(self.api, name) for name in self._list_projects()]

    def get(self, name: str = 'mindsdb') -> Project:
        """
        Get Project by name

        :param name: name of project
        :return: Project object
        """
        if name not in self._list_projects():
            raise AttributeError("Project doesn't exist")
        return Project(self.api, name)

    def create(self, name: str) -> Project:
        """
        Create new project and return it

        :param name: name of the project
        :return: Project object
        """

        ast_query = CreateDatabase(
            name=Identifier(name),
            engine='mindsdb',
            parameters={}
        )

        self.api.sql_query(ast_query.to_string())
        return Project(self.api, name)

    def drop(self, name: str):
        """
        Drop project from server

        :param name: name of the project
        """
        ast_query = DropDatabase(name=Identifier(name))
        self.api.sql_query(ast_query.to_string())


from typing import List

import pandas as pd

from mindsdb_sdk.databases import Databases


N_ROWS = 10


def get_dataframe_schema(df: pd.DataFrame):
    """
    Get the schema of a DataFrame

    :param df: DataFrame

    :return: list of dictionaries containing column names and types
    """
    # Get the dtypes Series
    try:
        df = df.convert_dtypes()
    except Exception as e:
        raise f"Error converting dtypes: {e}"

    dtypes = df.dtypes

    # Convert the dtypes Series into a list of dictionaries
    schema = [{"name": column, "type": dtype.name} for column, dtype in dtypes.items()]

    return schema


def get_table_schemas(database: Databases, included_tables: List[str] = None, n_rows: int = N_ROWS) -> dict:
    """
    Get table schemas from a database

    :param database: database object
    :param included_tables: list of table names to get schemas for
    :param n_rows: number of rows to fetch from each table

    :return: dictionary containing table schemas
    """

    tables = [table.name for table in database.tables.list()]

    if included_tables:
        tables = [table for table in tables if table in included_tables]

    table_schemas = {}
    for table in tables:
        table_df = database.get_table(table).limit(n_rows).fetch()
        table_schemas[table] = get_dataframe_schema(table_df)

    return table_schemas



import json
from logging import getLogger
from typing import List

import openai
from openai.types.chat import ChatCompletionToolChoiceOptionParam

from mindsdb_sdk.databases import Database
from tenacity import retry, wait_random_exponential, stop_after_attempt


DEFAULT_RETRY_MULTIPLIER = 1
DEFAULT_MAX_WAIT = 40
DEFAULT_STOP_AFTER_ATTEMPT = 3

logger = getLogger(__name__)


@retry(wait=wait_random_exponential(multiplier=DEFAULT_RETRY_MULTIPLIER, max=DEFAULT_MAX_WAIT), stop=stop_after_attempt(
    DEFAULT_RETRY_MULTIPLIER
))
def chat_completion_request(
        client: openai.OpenAI,
        model: str,
        messages: List[dict],
        tools: List = None,
        tool_choice: ChatCompletionToolChoiceOptionParam = None
):
    try:
        response = client.chat.completions.create(
            model=model,
            messages=messages,
            tools=tools,
            tool_choice=tool_choice,
        )
        return response
    except Exception as e:
        logger.warning("Unable to generate ChatCompletion response")
        logger.warning(f"Exception: {e}")
        return e


def make_openai_tool(function: callable, description: str = None) -> dict:
    """
    Make a generic OpenAI tool for a function

    :param function: function to generate metadata for
    :param description: description of the function

    :return: dictionary containing function metadata
    """

    import inspect
    import docstring_parser

    params = inspect.signature(function).parameters
    docstring = docstring_parser.parse(function.__doc__)

    # Get the first line of the docstring as the function description or use the user-provided description
    function_description = description or docstring.short_description

    function_dict = {
        "type":"function",
        "function":{
            "name":function.__name__,
            "description":function_description,
            "parameters":{
                "type":"object",
                "properties":{},
                "required":[]
            }
        }
    }

    for name, param in params.items():
        param_description = next((p.description for p in docstring.params if p.arg_name == name), '')

        # convert annotation type to string
        if param.annotation is not inspect.Parameter.empty:
            if inspect.isclass(param.annotation):
                param_type = param.annotation.__name__
            else:
                param_type = str(param.annotation)
        else:
            param_type = None

        function_dict["function"]["parameters"]["properties"][name] = {
            "type":param_type,
            "description":param_description
        }

        # Check if parameter is required
        if param.default == inspect.Parameter.empty:
            function_dict["function"]["parameters"]["required"].append(name)

    return function_dict


def make_query_tool(schema: dict) -> dict:
    """
    Make an OpenAI tool for querying a database connection in MindsDB

    :param schema: database schema

    :return: dictionary containing function metadata for openai tools
    """
    return {
        "type":"function",
        "function":{
            "name":"query_database",
            "description":"Use this function to answer user questions. Input should be a fully formed SQL query.",
            "parameters":{
                "type":"object",
                "properties":{
                    "query":{
                        "type":"string",
                        "description":f"""
                                    SQL query extracting info to answer the user's question.
                                    SQL should be written using this database schema:
                                    {schema}
                                    The query should be returned in plain text, not in JSON.
                                    """,
                    }
                },
                "required":["query"],
            },
        }
    }


def make_data_tool(
    model: str,
    data_source: str,
    description: str,
    connection_args: dict
):
    """
    tool passing mindsdb database connection details for datasource to litellm callback

    :param model: model name for text to sql completion
    :param data_source: data source name
    :param description: description of the data source
    :param connection_args: connection arguments for the data source

    :return: dictionary containing function metadata for openai tools
    """
    # Convert the connection_args dictionary to a JSON object
    connection_args_json = json.dumps(connection_args)

    tool_description = f"""
Queries the provided data source about user data. When calling this function, ALWAYS use the following arguments:
- model: {model}
- connection_args: {connection_args_json}
- data_source: {data_source}
- description: {description}
"""

    return {
        "type":"function",
        "function":{
            "name":"get_mindsdb_text_to_sql_completion",
            "description":tool_description,
            "parameters":{
                "type":"object",
                "properties":{
                    "model":{
                        "type":"string",
                        "description":"llm model name to use for text to sql completion",
                    },
                    "data_source":{
                        "type":"string",
                        "description":"Data source name",
                    },
                    "connection_args":{
                        "type":"string",
                        "description":"Connection arguments for the data source",
                    },
                    "description":{
                        "type":"string",
                        "description":"Description of the data source",
                    }
                },
                "required": ['data_source', 'connection_args', 'model', 'description']
            }
        }
    }


def extract_sql_query(result: str) -> str:
    """
    Extract the SQL query from an openai result string

    :param result: OpenAI result string
    :return: SQL query string
    """
    # Split the result into lines
    lines = result.split('\n')

    # Initialize an empty string to hold the query
    query = ""

    # Initialize a flag to indicate whether we're currently reading the query
    reading_query = False

    # Iterate over the lines
    for line in lines:
        # If the line starts with "SQLQuery:", start reading the query
        if line.startswith("SQLQuery:"):
            query = line[len("SQLQuery:"):].strip()
            reading_query = True
        # If the line starts with "SQLResult:", stop reading the query
        elif line.startswith("SQLResult:"):
            break
        # If we're currently reading the query, append the line to the query
        elif reading_query:
            query += " " + line.strip()

    # If no line starts with "SQLQuery:", return None
    if query == "":
        return None

    return query


def query_database(database: Database, query: str) -> str:
    """
    Execute a query on a database connection

    :param database: mindsdb Database object
    :param query: SQL query string

    :return: query results as a string
    """
    try:
        results = str(
            database.query(query).fetch()
        )
    except Exception as e:
        results = f"query failed with error: {e}"
    return results


def execute_function_call(message, database: Database = None) -> str:
    """
    Execute a function call in a message

    """
    if message.tool_calls[0].function.name == "query_database":
        query = json.loads(message.tool_calls[0].function.arguments)["query"]
        results = query_database(database, query)
    else:
        results = f"Error: function {message.tool_calls[0].function.name} does not exist"
    return results


def pretty_print_conversation(messages):
    # you will need to pip install termcolor
    from termcolor import colored
    role_to_color = {
        "system":"red",
        "user":"green",
        "assistant":"blue",
        "function":"magenta",
    }

    for message in messages:
        if message["role"] == "system":
            logger.info(colored(f"system: {message['content']}\n", role_to_color[message["role"]]))
        elif message["role"] == "user":
            logger.info(colored(f"user: {message['content']}\n", role_to_color[message["role"]]))
        elif message["role"] == "assistant" and message.get("function_call"):
            logger.info(colored(f"assistant: {message['function_call']}\n", role_to_color[message["role"]]))
        elif message["role"] == "assistant" and not message.get("function_call"):
            logger.info(colored(f"assistant: {message['content']}\n", role_to_color[message["role"]]))
        elif message["role"] == "function":
            logger.info(colored(f"function ({message['name']}): {message['content']}\n", role_to_color[message["role"]]))


from .agents import Agents
from .databases import Databases
from .projects import Project, Projects
from .ml_engines import MLEngines
from .handlers import Handlers
from .skills import Skills


class Server(Project):
    """
    Server instance allows to manipulate project and databases (integration) on mindsdb server

    Attributes for accessing to different objects:

        - projects, see :func:`~mindsdb_sdk.projects.Projects`
        - databases, see :func:`~mindsdb_sdk.databases.Databases`
        - ml_engines, see :func:`~mindsdb_sdk.ml_engines.MLEngines`
        - ml_handlers, see :func:`~mindsdb_sdk.handlers.MLHandlers`
        - data_handlers, see :func:`~mindsdb_sdk.handlers.DataHandlers`

    Server is also root(mindsdb) project and has attributes of project:
        - models, see :func:`~mindsdb_sdk.models.Models`
        - views, see :func:`~mindsdb_sdk.views.Views`
        - jobs, see :func:`~mindsdb_sdk.jobs.Jobs`

    """

    def __init__(self, api, skills: Skills = None, agents: Agents = None):
        # server is also mindsdb project
        project_name = 'mindsdb'
        self.databases = Databases(api)
        self.ml_engines = MLEngines(api)
        super().__init__(api, project_name, skills=skills, agents=agents, databases=self.databases, ml_engines=self.ml_engines)

        self.projects = Projects(api)

        # old api
        self.get_project = self.projects.get
        self.list_projects = self.projects.list
        self.create_project = self.projects.create
        self.drop_project = self.projects.drop


        # old api
        self.get_database = self.databases.get
        self.list_databases = self.databases.list
        self.create_database = self.databases.create
        self.drop_database = self.databases.drop


        self.ml_handlers = Handlers(self.api, 'ml')
        self.data_handlers = Handlers(self.api, 'data')

    def status(self) -> dict:
        """
        Get server information. It could content version
        Example of getting version for local:

        >>> print(server.status()['mindsdb_version'])

        :return: server status info
        """
        return self.api.status()

    def __repr__(self):
        return f'{self.__class__.__name__}({self.api.url})'





from typing import List

from mindsdb_sdk.utils.objects_collection import CollectionBase


class Skill():
    """Represents a MindsDB skill

    Working with skills:

    Get a skill by name:

    >>> skill = skills.get('my_skill')

    List all skills:

    >>> skills = skills.list()

    Create a new SQL skill:

    >>> text_to_sql_skill = skills.create('text_to_sql', 'sql', { 'tables': ['my_table'], 'database': 'my_database', 'description': 'my_description'})

    Update a skill:

    >>> skill.params = { 'tables': ['new_table'], 'database': 'new_database' }
    >>> updated_skill = skills.update('my_skill', skill)

    Delete a skill by name

    >>> skills.delete('my_skill')
    """
    def __init__(
            self,
            name: str,
            type: str,
            params: dict = None):
        self.name = name
        self.type = type
        self.params = params or {}

    def __eq__(self, other):
        if self.name != other.name:
            return False
        if self.type != other.type:
            return False
        return self.params == other.params

    def __repr__(self):
        return f'{self.__class__.__name__}(name: {self.name})'

    @classmethod
    def from_json(cls, json: dict):
        name = json['name']
        type = json['type']
        params = json['params']
        if json['type'] == 'sql':
            return SQLSkill(name, params['tables'], params['database'], params.get('description', ''))
        if json['type'] == 'retrieval':
            return RetrievalSkill(name, params['source'], params.get('description', ''))
        return Skill(name, type, params)


class SQLSkill(Skill):
    """Represents a MindsDB skill for agents to interact with MindsDB databases"""
    def __init__(self, name: str, tables: List[str], database: str, description: str):
        params = {
            'database': database,
            'tables': tables,
            'description': description
        }
        super().__init__(name, 'sql', params)

class RetrievalSkill(Skill):
    """Represents a MindsDB skill for agents to interact with MindsDB data sources"""
    def __init__(self, name: str, knowledge_base: str, description: str):
        params = {
            'source': knowledge_base,
            'description': description
        }
        super().__init__(name, 'retrieval', params)


class Skills(CollectionBase):
    """Collection for skills"""
    def __init__(self, api, project: str):
        self.api = api
        self.project = project

    def list(self) -> List[Skill]:
        """
        List available skills.

        :return: list of skills
        """
        data = self.api.skills(self.project)
        return [Skill.from_json(skill) for skill in data]

    def get(self, name: str) -> Skill:
        """
        Gets a skill by name.

        :param name: name of the skill

        :return: skill with the given name
        """
        data = self.api.skill(self.project, name)
        return Skill.from_json(data)

    def create(self, name: str, type: str, params: dict = None) -> Skill:
        """
        Create new skill and return it

        :param name: Name of the skill to be created
        :param type: Type of the skill to be created
        :param params: Parameters for the skill to be created

        :return: created skill object
        """
        _ = self.api.create_skill(self.project, name, type, params)
        if type == 'sql':
            return SQLSkill(name, params['tables'], params['database'], params['description'])
        return Skill(name, type, params)

    def update(self, name: str, updated_skill: Skill) -> Skill:
        """
        Update a skill by name.

        param name: Name of the skill to be updated
        :param updated_skill: Skill with updated fields

        :return: updated skillobject
        """
        data = self.api.update_skill(self.project, name, updated_skill.name, updated_skill.type, updated_skill.params)
        return Skill.from_json(data)

    def drop(self, name: str):
        """
        Drop a skill by name.

        :param name: Name of the skill to be dropped
        """
        _ = self.api.delete_skill(self.project, name)



import pandas as pd


class Query:
    def __init__(self, api, sql, database=None):
        self.api = api

        self.sql = sql
        self.database = database

    def __repr__(self):
        sql = self.sql.replace('\n', ' ')
        if len(sql) > 40:
            sql = sql[:37] + '...'

        return f'{self.__class__.__name__}({sql})'

    def fetch(self) -> pd.DataFrame:
        """
        Executes query in mindsdb server and returns result
        :return: dataframe with result
        """
        return self.api.sql_query(self.sql, self.database)

